
package org.objectweb.fractal.cecilia.adl;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;

import org.objectweb.fractal.adl.ADLErrors;
import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Compiler;
import org.objectweb.fractal.adl.CompilerError;
import org.objectweb.fractal.adl.ComponentVisitor;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.DefinitionVisitor;
import org.objectweb.fractal.adl.Factory;
import org.objectweb.fractal.adl.Loader;
import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.error.GenericErrors;
import org.objectweb.fractal.adl.util.FractalADLLogManager;
import org.objectweb.fractal.adl.xml.XMLNodeFactory;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.cecilia.adl.file.FileCollectionProvider;
import org.objectweb.fractal.cecilia.adl.file.FileProvider;
import org.objectweb.fractal.comete.adl.DeployableDefinition;
import org.objectweb.fractal.comete.adl.components.ExtendedDeployableComposite;
import org.objectweb.fractal.task.core.Record;
import org.objectweb.fractal.task.core.TaskException;
import org.objectweb.fractal.task.core.TaskInterface;
import org.objectweb.fractal.task.core.control.TaskExecutionException;

/**
 * Implementation of the {@link Factory} interface.
 * 
 * @see org.objectweb.fractal.adl.BasicFactory
 */
public class CometeLoaderFactory extends CeciliaFactory {

  /* PUBLIC METHODS */

  /** The name of the {@link Loader} client interface of this component. */
  public static final String LOADER_ITF_NAME        = "loader";

  /**
   * The name of the {@link Loader} client interface of this component for
   * deployable components.
   */
  public static final String DEPLOY_LOADER_ITF_NAME = "deploy-loader";

  /** The name of the {@link Compiler} client interface of this component. */
  public static final String VISITOR_ITF_NAME       = "visitor";

  /** The name of the {@link #nodeFactoryItf} client interface. */
  public static final String NODE_FACTORY_ITF_NAME  = "node-factory";

  
  // TODO this interface should not be redefined but it is private in CeciliaFactory
  /** The {@link Loader} used by this factory. */
  public Loader              loaderItf;

  /** The {@link Loader} used by this factory for deployable components. */
  public Loader              deployLoaderItf;

  /** The {@link ComponentVisitor} used by this factory. */
  private DefinitionVisitor  visitorItf;

  /** The {@link XMLNodeFactory} client interface. */
  public XMLNodeFactory      nodeFactoryItf;

  Map<Object, Object> cloneContext(final Map<Object, Object> context) {
    final Map<Object, Object> newContext = new HashMap<Object, Object>();

    newContext.putAll(context);

    return newContext;
  }

  @Override
  @SuppressWarnings("unchecked")
  public Object newComponent(final String name, final Map context)
      throws ADLException {

    final String soDir = (String) context.get("executable-name")
        + File.separator;
    final String buildDirectory = ((File) context.get("adlBuildDirectory"))
        .getParent()
        + File.separator;

    /* Load the fractal file using extensions for deployment */
    final Definition d = deployLoaderItf.load(name, context);

    // If the loaded definition is a deployable one, it means we need to compile
    // several components, stubs and skeletons, loader ...
    if (d instanceof DeployableDefinition) {

      final ExtendedDeployableComposite deploymentComposite = new ExtendedDeployableComposite(
          (ComponentContainer) d);

      deploymentComposite.genSources(context);

      deploymentComposite.compile(this, context, soDir, buildDirectory);

      return null;
    } else {
      // Otherwise, we just need to compile one component with the classic way
      return compile(d, context);
    }
  }

  // TODO check if we can avoid the following duplication by modifying Cecilia
  
  /*
   * This method is just copied from Cecilia. As we need to load the ADL to
   * known how it must be compiled, we need to be able compiling a Definition to avoid reloading 
   * the ADL again
   */
  protected Object newCeciliaComponent(final String name,
      final Map<Object, Object> context) throws ADLException {

    if (FractalADLLogManager.STEP.isLoggable(Level.FINE))
      FractalADLLogManager.STEP.fine("[basic-factory] Load ADL \'" + name
          + "\'");

    final Definition d = loaderItf.load(name, context);
    FractalADLLogManager.STEP
        .fine("[basic-factory] ADL loaded successfully, build task graph");

    return compile(d, context);
  }

  /*
   * This method is just copied from Cecilia. As we need to load the ADL to
   * known how it must be compiled, we need to be able compiling a Definition to avoid reloading 
   * the ADL again
   */
  protected Object compile(final Definition d, final Map<Object, Object> context)
      throws ADLException {

    Component taskGraph;
    try {
      taskGraph = visitorItf.visit(new ArrayList<Node>(), d, context);
    } catch (final TaskException e1) {
      throw new CompilerError(GenericErrors.INTERNAL_ERROR, e1,
          "Task component instantiation problem during the creation of the task graph.");
    }

    FractalADLLogManager.STEP
        .fine("[basic-factory] Task graph built successfully, execute tasks");

    Object result = null;
    try {
      final Record outputRecord = new Record("role:outputFiles, id:%", d);
      for (final Object taskItf : taskGraph.getFcInterfaces()) {
        if (outputRecord.equals(((TaskInterface) taskItf).getFcRecord())) {
          if (taskItf instanceof FileProvider)
            result = ((FileProvider) taskItf).getFile();
          if (taskItf instanceof FileCollectionProvider)
            result = ((FileCollectionProvider) taskItf).getFiles();
        }
      }
    } catch (final TaskExecutionException e) {
      if (e.getCause() instanceof ADLException)
        throw (ADLException) e.getCause();
      else
        throw new ADLException(ADLErrors.TASK_EXECUTION_ERROR, e.getCause());
    } catch (final Exception e) {
      throw new ADLException(ADLErrors.TASK_EXECUTION_ERROR, e);
    }

    FractalADLLogManager.STEP
        .fine("[basic-factory] Tasks executed successfully.");
    return result;
  }

  /**
   * Compile the given ADL.
   * 
   * @param adlName the name of the the name of the architecture (ADL) to
   *          compile.
   * @param context the Map object that contains the configuration of the
   *          compilation
   * @param execName the name of the executable file to be created. If it is
   *          null the default one (in context) is used.
   * @param buildDir the build directory used for temporary files. If it is
   *          null, the default one (in context) is used.
   * @return the list of objects that are the results of the compilation.
   * @throws ADLException if the compilation fails.
   */
  public Object compile(final String adlName,
      final Map<Object, Object> context, final String execName,
      final String buildDir) throws ADLException {

    // A new context must be created for each compilitation as it will modify it
    final Map<Object, Object> clonedContext = cloneContext(context);

    // Set the specified paths in the context
    if (execName != null) clonedContext.put("executable-name", execName);

    if (buildDir != null) {
      File adlBuildDirectory = (File) clonedContext.get("adlBuildDirectory");
      adlBuildDirectory = new File(buildDir + File.separator + "adl");
      clonedContext.put("adlBuildDirectory", adlBuildDirectory);

      // TODO changing the IDL directory does not work
      // File idlBuildDirectory = (File) clonedContext.get("idlBuildDirectory");
      // idlBuildDirectory = new File(buildDir + File.separator + "idl");
      // clonedContext.put("idlBuildDirectory", idlBuildDirectory);

      File objBuildDirectory = (File) clonedContext.get("objBuildDirectory");
      objBuildDirectory = new File(buildDir + File.separator + "obj");
      clonedContext.put("objBuildDirectory", objBuildDirectory);
    }

    return newCeciliaComponent(adlName, clonedContext);
  }

  // ------------------------------------------------------------------------
  // Implementation of the BindingController interface
  // ------------------------------------------------------------------------

  public void bindFc(final String itfName, final Object value)
      throws NoSuchInterfaceException, IllegalBindingException,
      IllegalLifeCycleException {

    if (itfName == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    try {
      if (itfName.equals(LOADER_ITF_NAME)) {
        this.loaderItf = (Loader) value;
      } else if (itfName.equals(DEPLOY_LOADER_ITF_NAME)) {
        this.deployLoaderItf = (Loader) value;
      } else if (itfName.equals(VISITOR_ITF_NAME)) {
        this.visitorItf = (DefinitionVisitor) value;
      } else if (itfName.equals(NODE_FACTORY_ITF_NAME)) {
        this.nodeFactoryItf = (XMLNodeFactory) value;
      } else {
        throw new NoSuchInterfaceException(
            "Unable to find the interface named '" + itfName + "'");
      }
    } catch (final ClassCastException cce) {
      cce.printStackTrace();
      throw new IllegalBindingException("Unmatching interface types");
    }
  }

  public String[] listFc() {
    return new String[]{LOADER_ITF_NAME, DEPLOY_LOADER_ITF_NAME,
        VISITOR_ITF_NAME, NODE_FACTORY_ITF_NAME};
  }

  public Object lookupFc(final String itfName) throws NoSuchInterfaceException {

    if (itfName == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itfName.equals(LOADER_ITF_NAME)) {
      return this.loaderItf;
    } else if (itfName.equals(DEPLOY_LOADER_ITF_NAME)) {
      return this.deployLoaderItf;
    } else if (itfName.equals(VISITOR_ITF_NAME)) {
      return this.visitorItf;
    } else if (itfName.equals(NODE_FACTORY_ITF_NAME)) {
      return this.nodeFactoryItf;
    } else {
      throw new NoSuchInterfaceException("Unable to find the interface named '"
          + itfName + "'");
    }
  }

  public void unbindFc(final String itfName) throws NoSuchInterfaceException,
      IllegalBindingException, IllegalLifeCycleException {

    if (itfName == null) {
      throw new IllegalArgumentException("Interface name can't be null");
    }

    if (itfName.equals(LOADER_ITF_NAME)) {
      this.loaderItf = null;
    } else if (itfName.equals(LOADER_ITF_NAME)) {
      this.deployLoaderItf = null;
    } else if (itfName.equals(VISITOR_ITF_NAME)) {
      this.visitorItf = null;
    } else if (itfName.equals(NODE_FACTORY_ITF_NAME)) {
      this.nodeFactoryItf = null;
    } else {
      throw new NoSuchInterfaceException("Unable to find the interface named '"
          + itfName + "'");
    }

  }
}
