
package org.objectweb.fractal.cecilia.improveast;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.bindings.Binding;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.components.ComponentDefinition;
import org.objectweb.fractal.adl.implementations.Implementation;
import org.objectweb.fractal.adl.implementations.ImplementationContainer;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;
import org.objectweb.fractal.adl.types.TypeInterface;
import org.objectweb.fractal.adl.types.TypeInterfaceUtil;

/**
 * A utilitary class to ease Cecilia AST manipulation
 */
public class AstUtil {

  /**
   * Get the parent of the given component
   * 
   * @param arg the component
   * @return the parent component
   */
  public static Node getParent(final Node arg) {
    final Object parent = arg.astGetDecoration("parent");
    if (parent instanceof Node) {
      return (Node) parent;
    } else
      return null;
  }

  /**
   * Get a sub-component from its name
   * 
   * @param parent the component in which the sub-component must be searched
   * @param name
   * @return the sub-component
   */
  public static Node getComponent(final Node parent, final String name) {
    if (name.equals("this")) return parent;
    if (parent instanceof ComponentContainer) {
      for (final Component c : ((ComponentContainer) parent).getComponents()) {
        if (c.getName().equals(name)) return c;
      }
    }
    return null;
  }

  /**
   * Get an interface from its name
   * 
   * @param parent the component from which the interface must be searched
   * @param name the name of the interface
   * @return the interface
   */
  public static Interface getInterface(final Node parent, final String name) {
    if (parent instanceof InterfaceContainer) {
      for (final Interface i : ((InterfaceContainer) parent).getInterfaces()) {
        if (TypeInterfaceUtil.isCollection(i)) {
          if (name.indexOf(i.getName()) == 0) return i;
        } else {
          if (i.getName().equals(name)) return i;
        }
      }
    }
    return null;
  }

  /**
   * Get the client interface of a binding
   * 
   * @param binding the binding
   * @return the client interface
   */
  public static Interface getClient(final Node binding) {
    if (binding instanceof Binding) {
      return (Interface) ((Binding) binding).astGetDecoration("client");
    }
    return null;
  }

  /**
   * Get the server interface of a binding
   * 
   * @param binding the binding
   * @return the server interface
   */
  public static Interface getServerItf(final Node binding) {
    if (binding instanceof Binding) {
      return (Interface) ((Binding) binding).astGetDecoration("server");
    }
    return null;
  }

  /**
   * Get the server component of a binding
   * 
   * @param binding the binding
   * @return the server component
   */
  // Suppress warning to cast Object to Set<Binding>
  @SuppressWarnings("unchecked")
  public static Set<ComponentContainer> getServerComponent(final Node binding) {
    if (binding instanceof Binding) {
      Set<ComponentContainer> comps = new HashSet<ComponentContainer>();
      Interface serverItf = (Interface) ((Binding) binding)
          .astGetDecoration("server");

      if (serverItf.astGetDecoration("bind") != null) {
        Set<Binding> bindings;
        if (!TypeInterfaceUtil.isCollection(serverItf)) {
          bindings = new HashSet<Binding>();
          bindings.add((Binding) serverItf.astGetDecoration("bind"));
        } else {
          bindings = (Set<Binding>) serverItf.astGetDecoration("bind");
        }
        for (final Binding nextBinding : bindings) {
          serverItf = (Interface) nextBinding.astGetDecoration("server");
          comps.addAll(getServerComponent(nextBinding));
        }
      } else {
        comps.add((ComponentContainer) getParent(serverItf));
      }

      return comps;
    }
    return null;
  }

  /**
   * Get the server interface bounded to the given client interface
   * 
   * @param itf the client interface
   * @return the server interface
   */
  public static Interface boundedServer(final Node itf) {
    if (itf instanceof Interface) {
      Interface i = (Interface) itf;
      if (!i.astGetDecorations().containsKey("bind")) {
        /* unbounded interface */
        return null;
      }
      do {
        final Binding b = (Binding) i.astGetDecoration("bind");
        i = (Interface) b.astGetDecoration("server");
      } while (i.astGetDecorations().containsKey("bind"));
      return i; /* return last bounded interface */
    }
    return null;
  }

  /**
   * Tell if the given component is a primitive
   * 
   * @param component the component
   * @return true if the component is a primitive
   */
  public static boolean isPrimitive(final Node component) {
    if (component instanceof ComponentContainer) {
      return ((ComponentContainer) component).getComponents().length == 0;
    }
    return false;
  }

  /**
   * Get the signature of the given interface
   * 
   * @param itf the interface
   * @return the signature
   */
  public static String getSignature(final Node itf) {
    if (itf instanceof TypeInterface) {
      return ((TypeInterface) itf).getSignature();
    }
    return null;
  }

  /**
   * Get the name of a component
   * 
   * @param node the component
   * @return the name
   */
  public static String getName(final Node node) {
    if (node instanceof Component) {
      return ((Component) node).getName();
    }
    return null;
  }

  /**
   * Get the names of all the parents of the given component
   * 
   * @param node the component
   * @return the names
   */
  public static LinkedList<String> getHierarchyNames(final Node node) {
    final LinkedList<String> l = new LinkedList<String>();
    Node n = node;
    while (n instanceof Component) {
      l.add(0, getName(n));
      n = getParent(n);
    }
    return l;
  }

  /**
   * Cast a node to an implementation node
   * 
   * @param node the node
   * @return the implementation node
   */
  public static Implementation toImplementation(final Node node) {
    if (node instanceof Implementation) {
      return (Implementation) node;
    }
    return null;
  }

  /**
   * Cast a node to an implementation container node
   * 
   * @param node the node
   * @return the implementation container node
   */
  public static ImplementationContainer toImplementationContainer(
      final Node node) {
    if (node instanceof ImplementationContainer) {
      return (ImplementationContainer) node;
    }
    return null;
  }

  /**
   * Cast a node to an component definition node
   * 
   * @param node the node
   * @return the component definition node
   */
  public static ComponentDefinition toComponentDefinition(final Node node) {
    if (node instanceof ComponentDefinition) {
      return (ComponentDefinition) node;
    }
    return null;
  }

  /**
   * Cast a node to an interface container node
   * 
   * @param node the node
   * @return the interface container node
   */
  public static InterfaceContainer toInterfaceContainer(final Node node) {
    if (node instanceof InterfaceContainer) {
      return (InterfaceContainer) node;
    }
    return null;
  }

  /**
   * Tell if the given interface is a client one
   * 
   * @param node the interface
   * @return true if it is a client interface
   */
  public static boolean isClientInterface(final Node node) {
    if (node instanceof TypeInterface) {
      return ((TypeInterface) node).getRole().equals("client");
    }
    return false;
  }

  /**
   * Tell if the given interface is a server one
   * 
   * @param node the interface
   * @return true if it is a server interface
   */
  public static boolean isServerInterface(final Node node) {
    if (node instanceof TypeInterface) {
      return ((TypeInterface) node).getRole().equals("server");
    }
    return false;
  }
}
