
package org.objectweb.fractal.cecilia.improveast;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.bindings.Binding;
import org.objectweb.fractal.adl.bindings.BindingContainer;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;
import org.objectweb.fractal.adl.types.TypeInterfaceUtil;

/**
 * Improve given AST. Recursive deep-first implementation.
 * <p>
 * The AST is enriched with the following decorations:
 * <ul>
 * <li>parent (for all nodes): give the parent of the node
 * <li>export (for server Interface nodes): give all the client Interface nodes
 * bounded to this one
 * <li>bind (for client Interface nodes): give the server Interface node
 * bounded to this one <lie>client (for Binding nodes): give the client
 * interface <lie>server (for Binding nodes): give the server interface
 * </ul>
 */
public class CeciliaImprover extends GenericImprover {

  /**
   * Cecilia Improver constructor. Registers AST improvers.
   */
  public CeciliaImprover() {
    /* register every improver */
    super();

    registerImprover(ComponentContainer.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return CeciliaImprover
            .improveComponentContainer((ComponentContainer) arg);
      }
    });

    registerImprover(InterfaceContainer.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return CeciliaImprover
            .improveInterfaceContainer((InterfaceContainer) arg);
      }
    });

    registerImprover(BindingContainer.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return CeciliaImprover.improveBindingContainer((BindingContainer) arg);
      }
    });

    registerImprover(Component.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return CeciliaImprover.improveComponent((Component) arg);
      }
    });

    registerImprover(Interface.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return CeciliaImprover.improveInterface((Interface) arg);
      }
    });

    registerImprover(Binding.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return CeciliaImprover.improveBinding((Binding) arg);
      }
    });
  }

  private static List<Node> improveComponentContainer(
      final ComponentContainer arg) {

    /* set sub-components parent */
    for (final Component c : arg.getComponents()) {
      c.astSetDecoration("parent", arg);
    }

    return null;
  }

  private static List<Node> improveInterfaceContainer(
      final InterfaceContainer arg) {

    /* set sub-interfaces parent */
    for (final Interface i : arg.getInterfaces()) {
      i.astSetDecoration("parent", arg);
      i.astSetDecoration("export", new HashSet<Binding>());
      if (TypeInterfaceUtil.isCollection(i)) {
        i.astSetDecoration("bind", new HashSet<Binding>());
      }
    }

    return null;
  }

  private static List<Node> improveBindingContainer(final BindingContainer arg) {

    /* set sub-bindings parent */
    for (final Binding b : arg.getBindings()) {
      b.astSetDecoration("parent", arg);
    }

    return null;
  }

  private static List<Node> improveComponent(final Component arg) {
    return null;
  }

  private static List<Node> improveInterface(final Interface arg) {
    return null;
  }

  // Suppress warning to cast Object to Set<Binding>
  @SuppressWarnings("unchecked")
  private static List<Node> improveBinding(final Binding arg) {

    /* improve binding iterfaces */
    Node parent;
    Node from, to;
    Interface ifrom, ito;

    parent = AstUtil.getParent(arg);

    // Add the server interface to the client interface decorations
    from = AstUtil.getComponent(parent, arg.getFrom().split("\\.")[0]);
    ifrom = AstUtil.getInterface(from, arg.getFrom().split("\\.")[1]);
    if (TypeInterfaceUtil.isCollection(ifrom)) {
      final Set<Binding> bs = (Set<Binding>) ifrom.astGetDecoration("bind");
      bs.add(arg);
    } else {
      ifrom.astSetDecoration("bind", arg);
    }

    // Add the client interface to the server interface decorations
    to = AstUtil.getComponent(parent, arg.getTo().split("\\.")[0]);
    ito = AstUtil.getInterface(to, arg.getTo().split("\\.")[1]);
    final Set<Binding> bs = (Set<Binding>) ito.astGetDecoration("export");
    bs.add(arg);

    arg.astSetDecoration("client", ifrom);
    arg.astSetDecoration("server", ito);

    return null;
  }

}
