
package org.objectweb.fractal.cecilia.improveast;

import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.objectweb.fractal.adl.Node;
import org.objectweb.fractal.adl.attributes.AttributesContainer;
import org.objectweb.fractal.adl.bindings.BindingContainer;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.interfaces.InterfaceContainer;

/**
 * This class is a generic improver that can be extended in order to add
 * information in the AST. If not extended, this one adds nothing
 */
public class GenericImprover {

  private final Map<String, List<Improver>> improvers;

  /**
   * Construct a default generic improver that adds nothing. This improver
   * should be extends to enrich the AST. *
   */
  public GenericImprover() {
    improvers = new HashMap<String, List<Improver>>();

    registerImprover(ComponentContainer.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return GenericImprover
            .improveComponentContainer((ComponentContainer) arg);
      }
    });

    registerImprover(InterfaceContainer.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return GenericImprover
            .improveInterfaceContainer((InterfaceContainer) arg);
      }
    });

    registerImprover(BindingContainer.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return GenericImprover.improveBindingContainer((BindingContainer) arg);
      }
    });

    registerImprover(AttributesContainer.class, new Improver() {
      public List<Node> improve(final Node arg) {
        return GenericImprover
            .improveAttributesContainer((AttributesContainer) arg);
      }
    });
  }

  /**
   * Apply all the registered improver on the given AST
   * @param arg the AST top node
   */
  public void improve(final Node arg) {
    final LinkedList<Node> q = new LinkedList<Node>();
    q.add(arg);
    while (!q.isEmpty()) {
      final Node n = q.remove();
      for (final String s : improvers.keySet()) {
        try {
          if (Class.forName(s).isInstance(n)) {
            for (final Improver i : improvers.get(s)) {
              final List<Node> l = i.improve(n);
              if (l != null) q.addAll(0, l);
            }
          }
        } catch (final ClassNotFoundException e) {
        }
      }
    }
  }

  /**
   * Register an improver
   * @param c the class on which this improver must be called
   * @param callback the callback that must be called for each object of the specified class
   */
  public void registerImprover(final Class c, final Improver callback) {
    final String name = c.getCanonicalName();
    if (!improvers.containsKey(name)) {
      improvers.put(name, new LinkedList<Improver>());
    }
    improvers.get(name).add(callback);
  }

  private static List<Node> improveComponentContainer(
      final ComponentContainer arg) {
    return Arrays.asList((Node[]) arg.getComponents());
  }

  private static List<Node> improveInterfaceContainer(
      final InterfaceContainer arg) {
    return Arrays.asList((Node[]) arg.getInterfaces());
  }

  private static List<Node> improveBindingContainer(final BindingContainer arg) {
    return Arrays.asList((Node[]) arg.getBindings());
  }

  private static List<Node> improveAttributesContainer(
      final AttributesContainer arg) {
    return Arrays.asList((Node) arg.getAttributes());
  }
}
