
package org.objectweb.fractal.comete.adl.components;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.adl.ADLException;
import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.bindings.Binding;
import org.objectweb.fractal.adl.bindings.BindingContainer;
import org.objectweb.fractal.adl.components.Component;
import org.objectweb.fractal.adl.components.ComponentContainer;
import org.objectweb.fractal.adl.interfaces.Interface;
import org.objectweb.fractal.adl.types.TypeInterfaceUtil;
import org.objectweb.fractal.cecilia.adl.CometeLoaderFactory;
import org.objectweb.fractal.cecilia.improveast.AstUtil;
import org.objectweb.fractal.comete.adl.DeployErrors;
import org.objectweb.fractal.comete.adl.bindings.DeployableBinding;
import org.objectweb.fractal.comete.adl.bindings.ExtendedDeployableBinding;
import org.objectweb.fractal.comete.gen.DeploymentCompositeStruct;
import org.objectweb.fractal.comete.gen.DeploymentSources;
import org.objectweb.fractal.comete.improveast.DeployAstUtil;

/**
 * Class for composite nodes of the deployment tree
 */
public class ExtendedDeployableComposite extends ExtendedDeployableComponent {

  /** Sub-components */
  Set<ExtendedDeployableComponent> subComponents = new HashSet<ExtendedDeployableComponent>();

  /** Bindings */
  Set<ExtendedDeployableBinding>   bindings      = new HashSet<ExtendedDeployableBinding>();

  /**
   * The constructor
   * 
   * @param cc the component container of this component
   * @throws ADLException
   */
  // Suppress warning to cast Object to Set<Binding>
  @SuppressWarnings("unchecked")
  public ExtendedDeployableComposite(final ComponentContainer cc)
      throws ADLException {
    super(cc);

    for (final Component comp : cc.getComponents()) {
      // Only process components that was defined with 'deployment' DTD,
      // as the others do not contain any deployment information
      // Don't process the system component as there is nothing to do for it
      if (comp instanceof DeployableComponent
          && !DeployAstUtil.isSystemComponent(comp)) {
        final DeployableComponent dc = (DeployableComponent) comp;

        ExtendedDeployableComponent newComp = null;
        if (DeployAstUtil.isDeployableComponent(dc)) {
          newComp = new ExtendedDeployablePrimitive(dc);
        } else {
          newComp = new ExtendedDeployableComposite(comp);
        }
        subComponents.add(newComp);

      }

    }

    if (subComponents.size() == 0) {
      // The composite has no primitive component, this means the processing
      // element was not specified
      throw new ADLException(DeployErrors.PE_MISSING, cc.astGetSource());
    }

    if (cc instanceof BindingContainer) {
      final BindingContainer bc = (BindingContainer) cc;
      for (final Binding binding : bc.getBindings()) {
        if (binding instanceof DeployableBinding) {

          final DeployableBinding db = (DeployableBinding) binding;

          if (db.getProtocol() == null) {
            throw new ADLException(DeployErrors.PROTOCOL_MISSING, db
                .astGetSource());
          }

          // Check this binding has the same protocol as the following one
          final Interface itf = AstUtil.getServerItf(binding);
          Set<Binding> bindings;
          if (!TypeInterfaceUtil.isCollection(itf)) {
            bindings = new HashSet<Binding>();
            bindings.add((Binding) itf.astGetDecoration("bind"));
          } else {
            bindings = (Set<Binding>) itf.astGetDecoration("bind");
          }
          for (final Binding nextBinding : bindings) {
            if (nextBinding != null && nextBinding instanceof DeployableBinding) {
              final DeployableBinding ndb = (DeployableBinding) nextBinding;
              if (ndb.getProtocol() != null
                  && !ndb.getProtocol().equals(db.getProtocol())) {
                throw new ADLException(DeployErrors.DIFFERENT_PROTOCOL, binding
                    .astGetSource(), nextBinding.astGetSource());
              }
            }
          }

          this.bindings.add(new ExtendedDeployableBinding(db));
        }
      }
    }
  }

  @Override
  public void genSources(final DeploymentSources sources,
      final String parentName) {

    String name = null;
    String fullName = null;

    // Get the name of the component
    if (componentContainer instanceof Definition) {
      name = ((Definition) componentContainer).getName();
    } else {
      name = ((DeployableComponent) componentContainer).getName();
    }

    String nextName;
    // Get the full name
    if (parentName == null)
      nextName = "";
    else if ("".equals(parentName))
      nextName = name;
    else
      nextName = parentName + "_" + name;

    if (parentName == null)
      fullName = "topComponent";
    else
      fullName = nextName;

    // Get the structure name
    if (DeployAstUtil.isPrimitive(componentContainer))
      // For primitive, the structure name is the definition as several
      // instances
      // of the same definition will load the same binary
      structName = (String) componentContainer.astGetDecoration("definition");
    else
      structName = fullName;

    // First get sources for sub-nodes
    for (final ExtendedDeployableComponent comp : this.subComponents) {
      comp.genSources(sources, nextName);
    }

    // Only add code for this component if it is not the top component or if we
    // must generate a loader
    if (parentName != null || !sources.hasLoader()) {
      // Create the structure for the composite
      final DeploymentCompositeStruct compositeStruct = new DeploymentCompositeStruct(
          fullName, structName);

      // Add sources for sub components
      for (final ExtendedDeployableComponent comp : this.subComponents) {
        comp.genSources(sources, nextName);

        final DeployableComponent dc = (DeployableComponent) comp.componentContainer;

        if (DeployAstUtil.isDeployableComponent(comp.componentContainer)) {
          // Each sub primitive has an entry into the composite structure
          final String subDefName = ((String) comp.componentContainer
              .astGetDecoration("definition")).replace(".", "_");
          compositeStruct.addSubPrimitive(dc.getName(), subDefName,
              comp.structName, dc.getStartOrder());
        } else {
          compositeStruct.addSubComposite(dc.getName(), null, comp.structName,
              dc.getStartOrder());
        }
      }

      // Add sources for bindings
      for (final ExtendedDeployableBinding binding : this.bindings) {
        final Interface fromItf = (Interface) binding.getBinding()
            .astGetDecoration("client");
        compositeStruct.addBinding(binding.getBinding().getFrom(), binding
            .getBinding().getTo(), binding.getBinding().getProtocol(),
            TypeInterfaceUtil.isCollection(fromItf));
      }

      // Register the component in the global sources
      sources.addComponentSource(structName, compositeStruct.getSource());
    }
  }

  @Override
  public Set<String> compile(final CometeLoaderFactory factory,
      final Map<Object, Object> context, final String soDir,
      final String buildDirectory, Set<String> definitionsDone)
      throws ADLException {

    super.compile(factory, context, soDir, buildDirectory, definitionsDone);

    for (final ExtendedDeployableComponent dc : subComponents) {
      definitionsDone = dc.compile(factory, context, soDir, buildDirectory,
          definitionsDone);
    }

    for (final ExtendedDeployableBinding db : bindings) {
      if (!DeployAstUtil.isSystemBinding(db.getBinding())) {
        definitionsDone = db.compile(factory, context, soDir, buildDirectory,
            definitionsDone);
      }
    }

    return definitionsDone;
  }

}
