
package org.objectweb.fractal.comete.gen;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.cecilia.adl.file.CodeWriter;

/**
 * Class for code generation of a composite C structure
 */
public class DeploymentCompositeStruct {

  class Binding {
    String fromCompName;
    String fromItfName;
    String toCompName;
    String toItfName;
    String type;

    /**
     * Constructor for a binding between sub-components of this component
     * 
     * @param fromCompName client component name
     * @param fromItfName client interface name
     * @param toCompName server component name
     * @param toItfName server interface name
     * @param type type of the binding (CBST, RPC, ..)
     */
    public Binding(final String fromCompName, final String fromItfName,
        final String toCompName, final String toItfName, final String type) {
      this.fromCompName = fromCompName;
      this.fromItfName = fromItfName;
      this.toCompName = toCompName;
      this.toItfName = toItfName;
      this.type = type;
    }
  }

  class Interface {
    String name;
    String subComponentName;
    String subItfName;
    String isCollection;

    /**
     * Constructor for a binding between a sub-component and an interface of the
     * composite
     * 
     * @param name interface name
     * @param subComponentName name of the sub-component bounded to this
     *            interface
     * @param subItfName name of the interface of the sub-component bounded to
     *            this interface
     * @param isCollection true if the interface is a collection one
     */
    public Interface(final String name, final String subComponentName,
        final String subItfName, final boolean isCollection) {
      this.name = name;
      this.subComponentName = subComponentName;
      this.subItfName = subItfName;
      if (isCollection == true) {
        this.isCollection = "1";
      } else {
        this.isCollection = "0";
      }
    }
  }

  class SubComponent {
    String  defName;
    String  instanceName;
    Integer startOrder;
    Integer index;
    String  structName;

    /**
     * Constructor for a sub-component of the composite
     * 
     * @param instanceName instance name of the sub-component
     * @param defName definition name of the sub-component
     * @param structName name of the C structure describing this sub-component
     * @param startOrder start order of the sub-component
     */
    public SubComponent(final String instanceName, final String defName,
        final String structName, final String startOrder) {
      this.defName = defName;
      this.structName = structName;
      this.instanceName = instanceName;
      if (startOrder != null)
        this.startOrder = Integer.parseInt(startOrder);
      else
        this.startOrder = 0;
    }
  }

  String            name            = null;
  String            structName      = null;
  Integer           nbSubPrimitives = 0;
  Set<SubComponent> subPrimitives   = new HashSet<SubComponent>();
  Integer           nbSubComposites = 0;
  Set<SubComponent> subComposites   = new HashSet<SubComponent>();
  Integer           nbBindings      = 0;
  Set<Binding>      bindings        = new HashSet<Binding>();
  Integer           nbItfs          = 0;
  Set<Interface>    itfs            = new HashSet<Interface>();

  /**
   * Constructor
   * 
   * @param name name of the composite
   * @param structName name of the C structure describing the composite
   */
  public DeploymentCompositeStruct(final String name, final String structName) {
    this.name = name;
    this.structName = structName;
  }

  /**
   * Add the given sub-primitive
   * 
   * @param instanceName instance name of the sub-component
   * @param defName definition name of the sub-component
   * @param structName name of the C structure describing the sub-component
   * @param startOrder start order of the sub-component
   */
  public void addSubPrimitive(final String instanceName, final String defName,
      final String structName, final String startOrder) {
    nbSubPrimitives++;
    subPrimitives.add(new SubComponent(instanceName, defName, structName,
        startOrder));
  }

  /**
   * Add the given sub-composite
   * 
   * @param instanceName instance name of the sub-component
   * @param defName definition name of the sub-component
   * @param structName name of the C structure describing the sub-component
   * @param startOrder start order of the sub-component
   */
  public void addSubComposite(final String instanceName, final String defName,
      final String structName, final String startOrder) {
    nbSubComposites++;
    subComposites.add(new SubComponent(instanceName, defName, structName,
        startOrder));
  }

  /**
   * Add the given binding
   * 
   * @param from field client in the ADL file
   * @param to field server in the ADL file
   * @param type type of the binding (CBST, RPC, ...)
   * @param isCollection true if the client interface is a collection one
   */
  public void addBinding(final String from, final String to, String type,
      final boolean isCollection) {
    final String fromCompName = from.split("\\.")[0];
    final String fromItfName = from.split("\\.")[1];

    final String toCompName = to.split("\\.")[0];
    final String toItfName = to.split("\\.")[1];

    if (fromCompName.equals("this")) {
      itfs.add(new Interface(fromItfName, toCompName, toItfName, isCollection));
      nbItfs++;
    } else if (toCompName.equals("this")) {
      itfs
          .add(new Interface(toItfName, fromCompName, fromItfName, isCollection));
      nbItfs++;
    } else {
      type = "comete_generic_api_ComponentManager_BIND_" + type.toUpperCase();

      bindings.add(new Binding(fromCompName, fromItfName, toCompName,
          toItfName, type));

      nbBindings++;
    }
  }

  /**
   * Generate the C code for this structure
   * 
   * @return the C code
   */
  public String getSource() {
    final CodeWriter cw = new CodeWriter();

    String subCompositesStructName = "0";
    String subPrimitivesStructName = "0";
    String bindingsStructName = "0";
    String itfsStructName = "0";
    String startOrderStructName = "0";

    // 
    final Map<String, Integer> subComponentsIndexes = new HashMap<String, Integer>();

    final SubComponent[] orderedSubPrimitives = new SubComponent[subPrimitives
        .size()];
    Integer currentIndex = 0;
    for (final SubComponent subComponent : subPrimitives) {
      orderedSubPrimitives[currentIndex] = subComponent;
      subComponent.index = currentIndex;
      subComponentsIndexes.put(subComponent.instanceName, currentIndex);
      currentIndex++;
    }

    final SubComponent[] orderedSubComposites = new SubComponent[subComposites
        .size()];
    currentIndex = 0;
    for (final SubComponent subComponent : subComposites) {
      orderedSubComposites[currentIndex] = subComponent;
      subComponent.index = currentIndex + subPrimitives.size();
      subComponentsIndexes.put(subComponent.instanceName, currentIndex
          + subPrimitives.size());
      currentIndex++;
    }

    // Create the list of component start order
    if (nbSubPrimitives != 0 || nbSubComposites != 0) {
      startOrderStructName = structName.replace(".", "_") + "_startOrder";
      cw.append("static int " + startOrderStructName + "[] = {");
      Integer currentOrder = 0;
      final Set<SubComponent> subComponents = new HashSet<SubComponent>();
      subComponents.addAll(subPrimitives);
      subComponents.addAll(subComposites);
      while (subComponents.size() != 0) {
        final Set<SubComponent> subComponentsClone = new HashSet<SubComponent>();
        subComponentsClone.addAll(subComponents);
        for (final SubComponent subComponent : subComponentsClone) {
          if (subComponent.startOrder == currentOrder) {
            cw.append(subComponent.index + ", ");
            subComponents.remove(subComponent);
          }
        }
        currentOrder++;
      }
      cw.appendln("};").endl();
    }

    // Create the array of bindings
    if (nbBindings != 0) {
      bindingsStructName = structName.replace(".", "_") + "_bindings";

      cw.appendln("static binding_t " + bindingsStructName + "[] = {");
      for (final Binding binding : bindings) {
        final String fromCompIndex = subComponentsIndexes.get(
            binding.fromCompName).toString();
        final String toCompIndex = subComponentsIndexes.get(binding.toCompName)
            .toString();
        cw.append(
            "{" + fromCompIndex + ", \"" + binding.fromItfName + "\", "
                + toCompIndex + ", \"" + binding.toItfName + "\", "
                + binding.type + "},").endl();
      }
      cw.appendln("};").endl();

    }

    // Create the array of interfaces
    if (nbItfs != 0) {
      itfsStructName = structName.replace(".", "_") + "_itfs";

      cw.appendln("itf_t " + itfsStructName + "[] = {");
      for (final Interface itf : itfs) {
        final String compIndex = subComponentsIndexes.get(itf.subComponentName)
            .toString();
        cw.append(
            "{\"" + itf.name + "\", " + compIndex + ", \"" + itf.subItfName
                + "\", " + itf.isCollection + "},").endl();
      }
      cw.appendln("};").endl();
    }

    if (nbSubPrimitives != 0) {
      subPrimitivesStructName = structName.replace(".", "_") + "_subPrimitives";

      // Create the array of pointers to the sub-components
      cw.appendln("static primitive_t *" + subPrimitivesStructName + "[] = {");

      for (final SubComponent subComponent : orderedSubPrimitives) {
        cw.appendln("&" + subComponent.structName + ", ");
      }
      cw.appendln("};").endl();

    }

    if (nbSubComposites != 0) {
      subCompositesStructName = structName.replace(".", "_") + "_subComposites";

      // Create the array of pointers to the sub-components
      cw.appendln("static composite_t *" + subCompositesStructName + "[] = {");

      for (final SubComponent subComponent : orderedSubComposites) {
        cw.appendln("&" + subComponent.structName + ", ");
      }
      cw.appendln("};").endl();

    }

    // Create the structure for the composite
    cw.appendln("static composite_t " + structName.replace(".", "_")
        + " = {{COMPOSITE, \"" + name + "\"}, " + nbSubPrimitives + ", "
        + nbSubComposites + ", " + subPrimitivesStructName + ", "
        + subCompositesStructName + ", " + startOrderStructName + ", "
        + nbBindings + ", " + bindingsStructName + ", " + nbItfs + ", "
        + itfsStructName + "};");
    cw.endl();

    return cw.toString();
  }
}
