
package org.objectweb.fractal.comete.gen;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.objectweb.fractal.adl.Definition;
import org.objectweb.fractal.adl.util.FractalADLLogManager;
import org.objectweb.fractal.cecilia.adl.file.CodeWriter;

/**
 * Class for code generation
 */
public class DeploymentSources {
  boolean     hasDeployer     = false;
  boolean     hasLoader       = false;
  String      deployerStructs = "";
  Set<String> composites      = new HashSet<String>();

  /**
   * Add C code for the deployer component
   * 
   * @param source C code
   */
  public void addDeployerSource(final String source) {
    hasDeployer = true;
    deployerStructs = deployerStructs.concat(source);
  }

  /** Add a loader */
  public void addLoader() {
    hasLoader = true;
  }

  /**
   * Tell if there is a loader
   * 
   * @return true if there is a loader
   */
  public boolean hasLoader() {
    return hasLoader;
  }

  /** 
   * Tell is there is a deployer
   * @return true if there is a deployer
   */
  public boolean hasDeployer() {
    return hasDeployer;
  }

  /**
   * Add C code for the given component
   * @param name component name
   * @param source component source
   */
  public void addComponentSource(final String name, final String source) {
    if (!composites.contains(name)) {
      composites.add(name);
      addDeployerSource(source);
    }
  }

  /**
   * Get the source of the deployer
   * @return C code of the deployer
   */
  public String getDeployerSource() {
    final CodeWriter cw = new CodeWriter();

    cw.appendln("/* This file is auto-generated */").endl();

    cw.append(deployerStructs);
    cw.endl();

    cw.appendln("static void *deployableComponents[] = {");
    for (final String composite : composites) {
      cw.append("(void *)&" + composite.replace(".", "_")).appendln(", ");
    }
    cw.appendln("0");
    cw.appendln("};").endl();

    cw.appendln("jint METHOD(lifecycle_controller, getFcState)(void* _this) {");
    cw.appendln("return fractal_api_ErrorConst_OPERATION_NOT_SUPPORTED;");
    cw.appendln("}").endl();

    cw.appendln("jint METHOD(lifecycle_controller, startFc) (void* _this) {");
    cw.appendln("DATA.deployableComponents = deployableComponents;");
    cw.appendln("return fractal_api_ErrorConst_OK;");
    cw.appendln("}").endl();

    cw.appendln("jint METHOD(lifecycle_controller, stopFc) (void* _this) {");
    cw.appendln("return fractal_api_ErrorConst_OK;");
    cw.appendln("}").endl();

    return cw.toString();
  }

  private void generateCometeLoaderContent(final Definition d, final Map context) {

    /* code writers */
    final CodeWriter hndl_cw = new CodeWriter();
    final CodeWriter code_cw = new CodeWriter();

    // TODO should not be generated
    
    /* comete handlers decalarations */
    hndl_cw.appendln("void *application;");
    code_cw
        .appendln("CALL(REQUIRED.deployer, bind, 0, \"allocator\", (void *)-1, \"allocator\", comete_generic_api_ComponentManager_BIND_RPC);");
    code_cw
        .appendln("CALL(REQUIRED.deployer, deploy, \"topComponent\", 0, &application);");
    code_cw.appendln("CALL(REQUIRED.deployer, start, application);");

    final CodeWriter cw = new CodeWriter();
    cw.appendln("#include <minus/bare/generic/debug/api/debug.h>").endl();
    // cw.appendln("#include <fractal/api/AttributeController.idl.h>").endl();
    cw.endl().appendln("DECLARE_DATA{};").endl();
    cw.appendln("#include <cecilia.h>");
    cw.appendln(
        "void METHOD(entry, main) (void *_this, jint argc, char **argv){")
        .endl();
    cw.append(hndl_cw.toString()).endl();
    cw.append(code_cw.toString()).endl();
    cw.appendln("}").endl();

    /* output file */
    try {
      final File outputDir = (File) context.get("adlBuildDirectory");
      final FileWriter output = new FileWriter(outputDir.getAbsolutePath()
          + File.separator + "loader.c");
      output.write(cw.toString());
      output.close();
    } catch (final IOException e) {
      FractalADLLogManager.STEP.info(e.toString());
    }
  }

  /**
   * Generate code for default loader
   * @param d ?
   * @param context ? 
   * @return ?
   */
  public String generateCometeLoader(final Definition d,
      final Map<Object, Object> context) {
    final StringBuffer xml = new StringBuffer("");

    generateCometeLoaderContent(d, context);

    // TODO should not be generated
    
    xml.append("<?xml version=\"1.0\" encoding=\"ISO-8859-1\" ?>\n");
    xml
        .append("<!DOCTYPE definition PUBLIC \"-//objectweb.org//DTD Fractal ADL 2.0//EN\"\n");
    xml
        .append("\"classpath://org/objectweb/fractal/cecilia/adl/parser/xml/cecilia.dtd\">\n");
    xml.append("<definition name=\"Loader\">\n");
    xml
        .append("<interface name=\"entry\" role=\"server\" signature=\"minus.common.generic.activity.api.Main\"/>\n");
    xml
        .append("<interface name=\"componentManager\" role=\"client\" signature=\"comete.generic.api.ComponentManager\"/>\n");
    xml
        .append("<interface name=\"allocator\" role=\"client\" signature=\"memory.api.Allocator\"/>\n");
    xml.append("<component name=\"loader\">\n");
    xml
        .append("<interface name=\"entry\" role=\"server\" signature=\"minus.common.generic.activity.api.Main\"/>\n");
    xml
        .append("<interface name=\"componentManager\" role=\"client\" signature=\"comete.generic.api.ComponentManager\"/>\n");
    xml
        .append("<interface name=\"deployer\" role=\"client\" signature=\"comete.Deployer\"/>\n");

    xml.append("<content class=\"loader\" language=\"thinkMC\"/>\n");
    xml.append("</component>\n");
    xml
        .append("<component name=\"deployer\" definition=\"comete.Deployer\"/>\n");
    xml
        .append("<binding client=\"loader.componentManager\" server=\"this.componentManager\"/>\n");
    xml
        .append("<binding client=\"loader.deployer\" server=\"deployer.deployer\"/>\n");
    xml
        .append("<binding client=\"deployer.componentManager\" server=\"this.componentManager\"/>\n");
    xml
        .append("<binding client=\"deployer.allocator\" server=\"this.allocator\"/>\n");
    xml.append("<binding client=\"this.entry\" server=\"loader.entry\"/>\n");
    xml.append("</definition>\n");

    /* output file */
    final File outputDir = (File) context.get("adlBuildDirectory");
    final String outputADL = "Loader";
    final String outputPath = outputDir.getAbsolutePath() + File.separator
        + outputADL + ".fractal";
    try {
      final FileWriter output = new FileWriter(outputPath);
      output.write(xml.toString());
      output.close();
    } catch (final IOException e) {
      FractalADLLogManager.STEP.info(e.toString());
    }

    URLClassLoader cl = (URLClassLoader) context.get("classloader");
    final URL newUrls[] = new URL[cl.getURLs().length + 1];
    System.arraycopy(cl.getURLs(), 0, newUrls, 1, cl.getURLs().length);
    try {
      newUrls[0] = outputDir.toURI().toURL();
    } catch (final MalformedURLException e) {
      // never append
      throw new Error(e);
    }

    cl = URLClassLoader.newInstance(newUrls);
    context.put("classloader", cl);

    return outputADL;
  }

}
